/*
 * JCuda - Java bindings for NVIDIA CUDA driver and runtime API
 *
 * Copyright (c) 2009 Marco Hutter - http://www.jcuda.org
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package jcuda.driver;

/**
 * Error codes.<br />
 * <br />
 * Most comments are taken from the CUDA reference manual.
 */
public class CUresult
{
    /**
     * The API call returned with no errors. In the case of query calls, this
     * can also mean that the operation being queried is complete (see
     * ::cuEventQuery() and ::cuStreamQuery()).
     */
    public static final int CUDA_SUCCESS                              = 0;

    /**
     * This indicates that one or more of the parameters passed to the API call
     * is not within an acceptable range of values.
     */
    public static final int CUDA_ERROR_INVALID_VALUE                  = 1;

    /**
     * The API call failed because it was unable to allocate enough memory to
     * perform the requested operation.
     */
    public static final int CUDA_ERROR_OUT_OF_MEMORY                  = 2;

    /**
     * This indicates that the CUDA driver has not been initialized with
     * ::cuInit() or that initialization has failed.
     */
    public static final int CUDA_ERROR_NOT_INITIALIZED                = 3;

    /**
     * This indicates that the CUDA driver is in the process of shutting down.
     */
    public static final int CUDA_ERROR_DEINITIALIZED                  = 4;


    /**
     * This indicates that no CUDA-capable devices were detected by the installed
     * CUDA driver.
     */
    public static final int CUDA_ERROR_NO_DEVICE                      = 100;

    /**
     * This indicates that the device ordinal supplied by the user does not
     * correspond to a valid CUDA device.
     */
    public static final int CUDA_ERROR_INVALID_DEVICE                 = 101;


    /**
     * This indicates that the device kernel image is invalid. This can also
     * indicate an invalid CUDA module.
     */
    public static final int CUDA_ERROR_INVALID_IMAGE                  = 200;

    /**
     * This most frequently indicates that there is no context bound to the
     * current thread. This can also be returned if the context passed to an
     * API call is not a valid handle (such as a context that has had
     * ::cuCtxDestroy() invoked on it). This can also be returned if a user
     * mixes different API versions (i.e. 3010 context with 3020 API calls).
     * See ::cuCtxGetApiVersion() for more details.
     */
    public static final int CUDA_ERROR_INVALID_CONTEXT                = 201;

    /**
     * This indicated that the context being supplied as a parameter to the
     * API call was already the active context.
     * \deprecated
     * This error return is deprecated as of CUDA 3.2. It is no longer an
     * error to attempt to push the active context via ::cuCtxPushCurrent().
     */
    public static final int CUDA_ERROR_CONTEXT_ALREADY_CURRENT        = 202;

    /**
     * This indicates that a map or register operation has failed.
     */
    public static final int CUDA_ERROR_MAP_FAILED                     = 205;

    /**
     * This indicates that an unmap or unregister operation has failed.
     */
    public static final int CUDA_ERROR_UNMAP_FAILED                   = 206;

    /**
     * This indicates that the specified array is currently mapped and thus
     * cannot be destroyed.
     */
    public static final int CUDA_ERROR_ARRAY_IS_MAPPED                = 207;

    /**
     * This indicates that the resource is already mapped.
     */
    public static final int CUDA_ERROR_ALREADY_MAPPED                 = 208;

    /**
     * This indicates that there is no kernel image available that is suitable
     * for the device. This can occur when a user specifies code generation
     * options for a particular CUDA source file that do not include the
     * corresponding device configuration.
     */
    public static final int CUDA_ERROR_NO_BINARY_FOR_GPU              = 209;

    /**
     * This indicates that a resource has already been acquired.
     */
    public static final int CUDA_ERROR_ALREADY_ACQUIRED               = 210;

    /**
     * This indicates that a resource is not mapped.
     */
    public static final int CUDA_ERROR_NOT_MAPPED                     = 211;

    /**
     * This indicates that a mapped resource is not available for access as an
     * array.
     */
    public static final int CUDA_ERROR_NOT_MAPPED_AS_ARRAY            = 212;

    /**
     * This indicates that a mapped resource is not available for access as a
     * pointer.
     */
    public static final int CUDA_ERROR_NOT_MAPPED_AS_POINTER          = 213;

    /**
     * This indicates that an uncorrectable ECC error was detected during
     * execution.
     */
    public static final int CUDA_ERROR_ECC_UNCORRECTABLE              = 214;

    /**
     * This indicates that the ::CUlimit passed to the API call is not
     * supported by the active device.
     */
    public static final int CUDA_ERROR_UNSUPPORTED_LIMIT              = 215;


    /**
     * This indicates that the device kernel source is invalid.
     */
    public static final int CUDA_ERROR_INVALID_SOURCE                 = 300;

    /**
     * This indicates that the file specified was not found.
     */
    public static final int CUDA_ERROR_FILE_NOT_FOUND                 = 301;

    /**
     * This indicates that a link to a shared object failed to resolve.
     */
    public static final int CUDA_ERROR_SHARED_OBJECT_SYMBOL_NOT_FOUND = 302;

    /**
     * This indicates that initialization of a shared object failed.
     */
    public static final int CUDA_ERROR_SHARED_OBJECT_INIT_FAILED      = 303;

    /**
     * This indicates that an OS call failed.
     */
    public static final int CUDA_ERROR_OPERATING_SYSTEM               = 304;


    /**
     * This indicates that a resource handle passed to the API call was not
     * valid. Resource handles are opaque types like ::CUstream and ::CUevent.
     */
    public static final int CUDA_ERROR_INVALID_HANDLE                 = 400;


    /**
     * This indicates that a named symbol was not found. Examples of symbols
     * are global/constant variable names, texture names, and surface names.
     */
    public static final int CUDA_ERROR_NOT_FOUND                      = 500;


    /**
     * This indicates that asynchronous operations issued previously have not
     * completed yet. This result is not actually an error, but must be indicated
     * differently than ::CUDA_SUCCESS (which indicates completion). Calls that
     * may return this value include ::cuEventQuery() and ::cuStreamQuery().
     */
    public static final int CUDA_ERROR_NOT_READY                      = 600;


    /**
     * An exception occurred on the device while executing a kernel. Common
     * causes include dereferencing an invalid device pointer and accessing
     * out of bounds shared memory. The context cannot be used, so it must
     * be destroyed (and a new one should be created). All existing device
     * memory allocations from this context are invalid and must be
     * reconstructed if the program is to continue using CUDA.
     */
    public static final int CUDA_ERROR_LAUNCH_FAILED                  = 700;

    /**
     * This indicates that a launch did not occur because it did not have
     * appropriate resources. This error usually indicates that the user has
     * attempted to pass too many arguments to the device kernel, or the
     * kernel launch specifies too many threads for the kernel's register
     * count. Passing arguments of the wrong size (i.e. a 64-bit pointer
     * when a 32-bit int is expected) is equivalent to passing too many
     * arguments and can also result in this error.
     */
    public static final int CUDA_ERROR_LAUNCH_OUT_OF_RESOURCES        = 701;

    /**
     * This indicates that the device kernel took too long to execute. This can
     * only occur if timeouts are enabled - see the device attribute
     * ::CU_DEVICE_ATTRIBUTE_KERNEL_EXEC_TIMEOUT for more information. The
     * context cannot be used (and must be destroyed similar to
     * ::CUDA_ERROR_LAUNCH_FAILED). All existing device memory allocations from
     * this context are invalid and must be reconstructed if the program is to
     * continue using CUDA.
     */
    public static final int CUDA_ERROR_LAUNCH_TIMEOUT                 = 702;

    /**
     * This error indicates a kernel launch that uses an incompatible texturing
     * mode.
     */
    public static final int CUDA_ERROR_LAUNCH_INCOMPATIBLE_TEXTURING  = 703;


    /**
     * This indicates that an unknown internal error has occurred.
     */
    public static final int CUDA_ERROR_UNKNOWN                        = 99;

    
    /**
     * Returns the String identifying the given CUresult
     *
     * @param result The CUresult value
     * @return The String identifying the given CUresult
     */
    public static String stringFor(int result)
    {
        switch (result)
        {
            case CUDA_SUCCESS : return "CUDA_SUCCESS";
            case CUDA_ERROR_INVALID_VALUE : return "CUDA_ERROR_INVALID_VALUE";
            case CUDA_ERROR_OUT_OF_MEMORY : return "CUDA_ERROR_OUT_OF_MEMORY";
            case CUDA_ERROR_NOT_INITIALIZED : return "CUDA_ERROR_NOT_INITIALIZED";
            case CUDA_ERROR_DEINITIALIZED : return "CUDA_ERROR_DEINITIALIZED";
            case CUDA_ERROR_NO_DEVICE : return "CUDA_ERROR_NO_DEVICE";
            case CUDA_ERROR_INVALID_DEVICE : return "CUDA_ERROR_INVALID_DEVICE";
            case CUDA_ERROR_INVALID_IMAGE : return "CUDA_ERROR_INVALID_IMAGE";
            case CUDA_ERROR_INVALID_CONTEXT : return "CUDA_ERROR_INVALID_CONTEXT";
            case CUDA_ERROR_CONTEXT_ALREADY_CURRENT : return "CUDA_ERROR_CONTEXT_ALREADY_CURRENT";
            case CUDA_ERROR_MAP_FAILED : return "CUDA_ERROR_MAP_FAILED";
            case CUDA_ERROR_UNMAP_FAILED : return "CUDA_ERROR_UNMAP_FAILED";
            case CUDA_ERROR_ARRAY_IS_MAPPED : return "CUDA_ERROR_ARRAY_IS_MAPPED";
            case CUDA_ERROR_ALREADY_MAPPED : return "CUDA_ERROR_ALREADY_MAPPED";
            case CUDA_ERROR_NO_BINARY_FOR_GPU : return "CUDA_ERROR_NO_BINARY_FOR_GPU";
            case CUDA_ERROR_ALREADY_ACQUIRED : return "CUDA_ERROR_ALREADY_ACQUIRED";
            case CUDA_ERROR_NOT_MAPPED : return "CUDA_ERROR_NOT_MAPPED";
            case CUDA_ERROR_NOT_MAPPED_AS_ARRAY : return "CUDA_ERROR_NOT_MAPPED_AS_ARRAY";
            case CUDA_ERROR_NOT_MAPPED_AS_POINTER : return "CUDA_ERROR_NOT_MAPPED_AS_POINTER";
            case CUDA_ERROR_ECC_UNCORRECTABLE : return "CUDA_ERROR_ECC_UNCORRECTABLE";
            case CUDA_ERROR_UNSUPPORTED_LIMIT : return "CUDA_ERROR_UNSUPPORTED_LIMIT";
            case CUDA_ERROR_INVALID_SOURCE : return "CUDA_ERROR_INVALID_SOURCE";
            case CUDA_ERROR_FILE_NOT_FOUND : return "CUDA_ERROR_FILE_NOT_FOUND";
            case CUDA_ERROR_SHARED_OBJECT_SYMBOL_NOT_FOUND : return "CUDA_ERROR_SHARED_OBJECT_SYMBOL_NOT_FOUND";
            case CUDA_ERROR_SHARED_OBJECT_INIT_FAILED : return "CUDA_ERROR_SHARED_OBJECT_INIT_FAILED";
            case CUDA_ERROR_INVALID_HANDLE : return "CUDA_ERROR_INVALID_HANDLE";
            case CUDA_ERROR_NOT_FOUND : return "CUDA_ERROR_NOT_FOUND";
            case CUDA_ERROR_NOT_READY : return "CUDA_ERROR_NOT_READY";
            case CUDA_ERROR_LAUNCH_FAILED : return "CUDA_ERROR_LAUNCH_FAILED";
            case CUDA_ERROR_LAUNCH_OUT_OF_RESOURCES : return "CUDA_ERROR_LAUNCH_OUT_OF_RESOURCES";
            case CUDA_ERROR_LAUNCH_TIMEOUT : return "CUDA_ERROR_LAUNCH_TIMEOUT";
            case CUDA_ERROR_LAUNCH_INCOMPATIBLE_TEXTURING : return "CUDA_ERROR_LAUNCH_INCOMPATIBLE_TEXTURING";
            case CUDA_ERROR_UNKNOWN : return "CUDA_ERROR_UNKNOWN";
        }
        return "INVALID CUresult: "+result;
    }

    /**
     * Private constructor to prevent instantiation.
     */
    private CUresult()
    {
    }

}
